"""
Copyright ©2017. The Regents of the University of California (Regents).
All Rights Reserved. Permission to use, copy, modify, and distribute this
software and its documentation for educational, research, and not-for-profit
purposes, without fee and without a signed licensing agreement, is hereby
granted, provided that the above copyright notice, this paragraph and the
following two paragraphs appear in all copies, modifications, and
distributions. Contact The Office of Technology Licensing, UC Berkeley,
2150 Shattuck Avenue, Suite 510, Berkeley, CA 94720-1620, (510) 643-7201,
otl@berkeley.edu, http://ipira.berkeley.edu/industry-info for commercial
licensing opportunities.

IN NO EVENT SHALL REGENTS BE LIABLE TO ANY PARTY FOR DIRECT, INDIRECT,
SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING LOST PROFITS,
ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF
REGENTS HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

REGENTS SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
PURPOSE. THE SOFTWARE AND ACCOMPANYING DOCUMENTATION, IF ANY, PROVIDED
HEREUNDER IS PROVIDED "AS IS". REGENTS HAS NO OBLIGATION TO PROVIDE
MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
"""
# Access levels
READ_ONLY_ACCESS = "READ_ONLY"
READ_WRITE_ACCESS = "READ_WRITE"
WRITE_ACCESS = "WRITE"

# Formatting
JSON_INDENT = 2

# Train / Test
TRAIN_ID = 0
TEST_ID = 1

# Extensions
COLOR_IMAGE_EXTS = [".png", ".jpg"]
INTR_EXTENSION = ".intr"

# Image constants
MIN_DEPTH = 0.25
MAX_DEPTH = 1.25
MAX_IR = 65535
METERS_TO_MM = 1000.0
MM_TO_METERS = 1.0 / METERS_TO_MM



"""
Classes for easy chessboard registration
Authors: Jeff Mahler and Jacky Liang
"""
import logging
import matplotlib.pyplot as plt
import numpy as np
import time

from .image import DepthImage
from .points import PointCloud, Point
from .rigid_transformations import RigidTransform


class ChessboardRegistrationResult(object):
    """Struct to encapsulate results of camera-to-chessboard registration.

    Attributes
    ----------
    T_camera_cb : :obj:`autolab_core.RigidTransform`
        transformation from camera to chessboard frame
    cb_points_camera : :obj:`autolab_core.PointCloud`
        3D locations of chessboard corners in the camera frame
    """

    def __init__(self, T_camera_cb, cb_points_camera):
        self.T_camera_cb = T_camera_cb
        self.cb_points_cam = cb_points_camera


class CameraChessboardRegistration:
    """
    Namespace for camera to chessboard registration functions.
    """

    @staticmethod
    def register(sensor, config):
        """
        Registers a camera to a chessboard.

        Parameters
        ----------
        sensor : :obj:`perception.RgbdSensor`
            the sensor to register
        config : :obj:`autolab_core.YamlConfig` or :obj:`dict`
            configuration file for registration

        Returns
        -------
        :obj:`ChessboardRegistrationResult`
            the result of registration

        Notes
        -----
        The config must have the parameters specified in the
        Other Parameters section.

        Other Parameters
        ----------------
        num_transform_avg : int
            the number of independent registrations to average together
        num_images : int
            the number of images to read for each independent registration
        corners_x : int
            the number of chessboard corners in the x-direction
        corners_y : int
            the number of chessboard corners in the y-direction
        color_image_rescale_factor : float
            amount to rescale the color image for detection
            (numbers around 4-8 are useful)
        vis : bool
            whether or not to visualize the registration
        """
        # read config
        num_transform_avg = config["num_transform_avg"]
        num_images = config["num_images"]
        sx = config["corners_x"]
        sy = config["corners_y"]
        point_order = config["point_order"]
        color_image_rescale_factor = config["color_image_rescale_factor"]
        flip_normal = config["flip_normal"]
        y_points_left = False
        if "y_points_left" in config.keys() and sx == sy:
            y_points_left = config["y_points_left"]
            num_images = 1
        vis = config["vis"]

        # read params from sensor
        logging.info("Registering camera %s" % (sensor.frame))

        # repeat registration multiple times and average results
        points_3d_plane = PointCloud(
            np.zeros([3, sx * sy]), frame=sensor.ir_frame
        )

        k = 0
        while k < num_transform_avg:
            # average a bunch of depth images together
            depth_ims = None
            for i in range(num_images):
                start = time.time()
                small_color_im, new_depth_im, _ = sensor.frames()
                end = time.time()
                logging.info("Frames Runtime: %.3f" % (end - start))
                if depth_ims is None:
                    depth_ims = np.zeros(
                        [new_depth_im.height, new_depth_im.width, num_images]
                    )
                depth_ims[:, :, i] = new_depth_im.data

            med_depth_im = np.median(depth_ims, axis=2)
            depth_im = DepthImage(med_depth_im, sensor.ir_frame)

            # find the corner pixels in an upsampled version of the color image
            big_color_im = small_color_im.resize(color_image_rescale_factor)
            corner_px = big_color_im.find_chessboard(sx=sx, sy=sy)

            # Visualize corner detections on big color image if vis==True
            if vis and corner_px is not None:
                plt.figure()
                plt.imshow(big_color_im.data)
                for i in range(sx):
                    plt.scatter(corner_px[i, 0], corner_px[i, 1], s=25, c="b")
                plt.show()

            elif corner_px is None:
                logging.error(
                    "No chessboard detected! Check camera exposure settings"
                )
                continue

            # convert back to original image
            small_corner_px = corner_px / color_image_rescale_factor

            if vis:
                plt.figure()
                plt.imshow(small_color_im.data)
                for i in range(sx):
                    plt.scatter(
                        small_corner_px[i, 0],
                        small_corner_px[i, 1],
                        s=25,
                        c="b",
                    )
                plt.axis("off")
                plt.show()

            # project points into 3D
            camera_intr = sensor.ir_intrinsics
            points_3d = camera_intr.deproject(depth_im)

            # get round chessboard ind
            corner_px_round = np.round(small_corner_px).astype(np.uint16)
            corner_ind = depth_im.ij_to_linear(
                corner_px_round[:, 0], corner_px_round[:, 1]
            )
            if corner_ind.shape[0] != sx * sy:
                logging.warning("Did not find all corners. Discarding...")
                continue

            # average 3d points
            points_3d_plane = (k * points_3d_plane + points_3d[corner_ind]) / (
                k + 1
            )
            logging.info(
                "Registration iteration %d of %d"
                % (k + 1, config["num_transform_avg"])
            )
            k += 1

        # fit a plane to the chessboard corners
        X = np.c_[
            points_3d_plane.x_coords,
            points_3d_plane.y_coords,
            np.ones(points_3d_plane.num_points),
        ]
        y = points_3d_plane.z_coords
        A = X.T.dot(X)
        b = X.T.dot(y)
        w = np.linalg.inv(A).dot(b)
        n = np.array([w[0], w[1], -1])
        n = n / np.linalg.norm(n)
        if flip_normal:
            n = -n
        mean_point_plane = points_3d_plane.mean()

        # find x-axis of the chessboard coordinates on the fitted plane
        T_camera_table = RigidTransform(
            translation=-points_3d_plane.mean().data,
            from_frame=points_3d_plane.frame,
            to_frame="table",
        )
        points_3d_centered = T_camera_table * points_3d_plane

        # get points along y
        if point_order == "row_major":
            coord_pos_x = int(np.floor(sx * sy / 2.0))
            coord_neg_x = int(np.ceil(sx * sy / 2.0))

            points_pos_x = points_3d_centered[coord_pos_x:]
            points_neg_x = points_3d_centered[:coord_neg_x]
            x_axis = np.mean(points_pos_x.data, axis=1) - np.mean(
                points_neg_x.data, axis=1
            )
            x_axis = x_axis - np.vdot(x_axis, n) * n
            x_axis = x_axis / np.linalg.norm(x_axis)
            y_axis = np.cross(n, x_axis)
        else:
            coord_pos_y = int(np.floor(sx * (sy - 1) / 2.0))
            coord_neg_y = int(np.ceil(sx * (sy + 1) / 2.0))
            points_pos_y = points_3d_centered[:coord_pos_y]
            points_neg_y = points_3d_centered[coord_neg_y:]
            y_axis = np.mean(points_pos_y.data, axis=1) - np.mean(
                points_neg_y.data, axis=1
            )
            y_axis = y_axis - np.vdot(y_axis, n) * n
            y_axis = y_axis / np.linalg.norm(y_axis)
            x_axis = np.cross(-n, y_axis)

        # produce translation and rotation from plane center and chessboard
        # basis
        rotation_cb_camera = RigidTransform.rotation_from_axes(
            x_axis, y_axis, n
        )
        translation_cb_camera = mean_point_plane.data
        T_cb_camera = RigidTransform(
            rotation=rotation_cb_camera,
            translation=translation_cb_camera,
            from_frame="cb",
            to_frame=sensor.frame,
        )

        if y_points_left and np.abs(T_cb_camera.y_axis[1]) > 0.1:
            if T_cb_camera.x_axis[0] > 0:
                T_cb_camera.rotation = T_cb_camera.rotation.dot(
                    RigidTransform.z_axis_rotation(-np.pi / 2).T
                )
            else:
                T_cb_camera.rotation = T_cb_camera.rotation.dot(
                    RigidTransform.z_axis_rotation(np.pi / 2).T
                )
        T_camera_cb = T_cb_camera.inverse()

        # optionally display cb corners with detected pose in 3d space
        if config["debug"]:
            # display image with axes overlayed
            cb_center_im = camera_intr.project(
                Point(T_cb_camera.translation, frame=sensor.ir_frame)
            )
            cb_x_im = camera_intr.project(
                Point(
                    T_cb_camera.translation
                    + T_cb_camera.x_axis * config["scale_amt"],
                    frame=sensor.ir_frame,
                )
            )
            cb_y_im = camera_intr.project(
                Point(
                    T_cb_camera.translation
                    + T_cb_camera.y_axis * config["scale_amt"],
                    frame=sensor.ir_frame,
                )
            )
            cb_z_im = camera_intr.project(
                Point(
                    T_cb_camera.translation
                    + T_cb_camera.z_axis * config["scale_amt"],
                    frame=sensor.ir_frame,
                )
            )
            x_line = np.array([cb_center_im.data, cb_x_im.data])
            y_line = np.array([cb_center_im.data, cb_y_im.data])
            z_line = np.array([cb_center_im.data, cb_z_im.data])

            plt.figure()
            plt.imshow(small_color_im.data)
            plt.scatter(cb_center_im.data[0], cb_center_im.data[1])
            plt.plot(x_line[:, 0], x_line[:, 1], c="r", linewidth=3)
            plt.plot(y_line[:, 0], y_line[:, 1], c="g", linewidth=3)
            plt.plot(z_line[:, 0], z_line[:, 1], c="b", linewidth=3)
            plt.axis("off")
            plt.title("Chessboard frame in camera %s" % (sensor.frame))
            plt.show()

        return ChessboardRegistrationResult(T_camera_cb, points_3d_plane)



"""
Classes for image detection
Author: Jeff Mahler
"""
from abc import ABCMeta, abstractmethod
import colorsys
import logging
import matplotlib.pyplot as plt
import numpy as np
import scipy.ndimage.morphology as snm

from .image import BinaryImage, DepthImage
from .primitives import Box


class RgbdDetection(object):
    """Struct to wrap the results of rgbd detection.

    Attributes
    ----------
    height : int
        height of detected object in pixels
    width : int
        width of detected object in pixels
    camera_intr : :obj:`CameraIntrinsics`
        camera intrinsics that project the detected window into
        the camera center
    query_im : :obj:`ColorImage`
        binary segmask for detected object as a 3-channel color image,
        for backwards comp
    color_im : :obj:`ColorImage`
        color thumbnail of detected object, with object masked
        if binary mask available
    depth_im : :obj:`DepthImage`
        depth thumbnail of detected object, with object masked
        if binary mask available
    binary_im : :obj:`BinaryImage`
        binary segmask of detected object
    depth_im_table : :obj:`DepthImage`
        unmasked depth image, for backwards comp
    cropped_ir_intrinsics : :obj:`CameraIntrinsics`
        alias for camera_intr, for backwards comp
    point_normal_cloud : :obj:`PointNormalCloud`
        point cloud with normals for the detected object
    """

    def __init__(
        self,
        color_thumbnail,
        depth_thumbnail,
        bounding_box,
        binary_thumbnail=None,
        camera_intr=None,
        contour=None,
    ):
        self.color_thumbnail = (
            color_thumbnail  # cropped color image from bounding box
        )
        self.depth_thumbnail = (
            depth_thumbnail  # cropped depth image from bounding box
        )
        self.bounding_box = (
            bounding_box  # bounding box in original source image
        )
        self.binary_thumbnail = (
            binary_thumbnail  # optional binary image masking the object
        )
        self.camera_intr = (
            camera_intr  # optional intrinsics of camera taking the thumbnail
        )
        self.contour = (
            contour  # optional contour describing the object boundary
        )

    @property
    def height(self):
        return self.bounding_box.dims[0]

    @property
    def width(self):
        return self.bounding_box.dims[1]

    @property
    def query_im(self):
        return self.binary_im.to_color()

    @property
    def color_im(self):
        if self.binary_thumbnail is None:
            return self.color_thumbnail
        return self.color_thumbnail.mask_binary(self.binary_thumbnail)

    @property
    def depth_im(self):
        if self.binary_thumbnail is None:
            return self.depth_thumbnail
        return self.depth_thumbnail.mask_binary(self.binary_thumbnail)

    @property
    def binary_im(self):
        return self.binary_thumbnail

    @property
    def depth_im_table(self):
        return self.depth_thumbnail

    @property
    def cropped_ir_intrinsics(self):
        return self.camera_intr

    @property
    def virtual_camera_intrinsics(self):
        return self.camera_intr

    @property
    def point_normal_cloud(self):
        if self.camera_intr is None:
            return None
        point_normal_cloud = self.depth_thumbnail.point_normal_cloud(
            self.camera_intr
        )
        point_normal_cloud.remove_zero_points()
        return point_normal_cloud


class RgbdDetector(object):
    """Wraps methods for as many distinct objects in the image as possible."""

    __metaclass__ = ABCMeta

    @abstractmethod
    def detect(
        self,
        color_im,
        depth_im,
        cfg,
        camera_intr=None,
        T_camera_world=None,
        segmask=None,
    ):
        """
        Detects all relevant objects in an rgbd image pair.

        Parameters
        ----------
        color_im : :obj:`ColorImage`
            color image for detection
        depth_im : :obj:`DepthImage`
            depth image for detection (corresponds to color image)
        cfg : :obj:`YamlConfig`
            parameters of detection function
        camera_intr : :obj:`CameraIntrinsics`
            intrinsics of the camera
        T_camera_world : :obj:`autolab_core.RigidTransform`
            registration of the camera to world frame
        segmask : :obj:`BinaryImage`
            optional segmask of invalid pixels

        Returns
        ------
        :obj:`list` of :obj:`RgbdDetection`
            all detections in the image
        """
        pass


class RgbdForegroundMaskDetector(RgbdDetector):
    """Detect by identifying all connected components in the foreground of
    the images using background subtraction.
    """

    def detect(
        self,
        color_im,
        depth_im,
        cfg,
        camera_intr=None,
        T_camera_world=None,
        segmask=None,
    ):
        """
        Detects all relevant objects in an rgbd image pair
        using foreground masking.

        Parameters
        ----------
        color_im : :obj:`ColorImage`
            color image for detection
        depth_im : :obj:`DepthImage`
            depth image for detection (corresponds to color image)
        cfg : :obj:`YamlConfig`
            parameters of detection function
        camera_intr : :obj:`CameraIntrinsics`
            intrinsics of the camera
        T_camera_world : :obj:`autolab_core.RigidTransform`
            registration of the camera to world frame
        segmask : :obj:`BinaryImage`
            optional segmask of invalid pixels

        Returns
        ------
        :obj:`list` of :obj:`RgbdDetection`
            all detections in the image
        """
        # read params
        foreground_mask_tolerance = cfg["foreground_mask_tolerance"]
        min_contour_area = cfg["min_contour_area"]
        max_contour_area = cfg["max_contour_area"]
        w = cfg["filter_dim"]

        # mask image using background detection
        bgmodel = color_im.background_model()
        binary_im = color_im.foreground_mask(
            foreground_mask_tolerance, bgmodel=bgmodel
        )

        # filter the image
        y, x = np.ogrid[-w / 2 + 1 : w / 2 + 1, -w / 2 + 1 : w / 2 + 1]
        mask = x * x + y * y <= w / 2 * w / 2
        filter_struct = np.zeros([w, w]).astype(np.uint8)
        filter_struct[mask] = 1
        binary_im_filtered = binary_im.apply(
            snm.grey_closing, structure=filter_struct
        )

        visualize = False
        if visualize:
            plt.figure()
            plt.imshow(binary_im_filtered.data, cmap=plt.cm.gray)
            plt.axis("off")
            plt.show()

        # find all contours
        contours = binary_im_filtered.find_contours(
            min_area=min_contour_area, max_area=max_contour_area
        )

        # convert contours to detections
        detections = []
        for contour in contours:
            box = contour.bounding_box
            color_thumbnail = color_im.crop(
                box.height, box.width, box.ci, box.cj
            )
            depth_thumbnail = depth_im.crop(
                box.height, box.width, box.ci, box.cj
            )
            binary_thumbnail = binary_im_filtered.crop(
                box.height, box.width, box.ci, box.cj
            )
            thumbnail_intr = camera_intr
            if camera_intr is not None:
                thumbnail_intr = camera_intr.crop(
                    box.height, box.width, box.ci, box.cj
                )
            detections.append(
                RgbdDetection(
                    color_thumbnail,
                    depth_thumbnail,
                    box,
                    binary_thumbnail=binary_thumbnail,
                    contour=contour,
                    camera_intr=thumbnail_intr,
                )
            )
        return detections


class RgbdForegroundMaskQueryImageDetector(RgbdDetector):
    """Detect by identifying all connected components in the foreground of
    the images using background subtraction. Converts all detections within
    a specified area into query images for a cnn. Optionally resegements
    the images using KMeans to remove spurious background pixels.
    """

    def _segment_color(
        self, color_im, bounding_box, bgmodel, cfg, vis_segmentation=False
    ):
        """Re-segments a color image to isolate an object of interest
        using foreground masking and kmeans"""
        # read params
        foreground_mask_tolerance = cfg["foreground_mask_tolerance"]
        color_seg_rgb_weight = cfg["color_seg_rgb_weight"]
        color_seg_num_clusters = cfg["color_seg_num_clusters"]
        color_seg_hsv_weight = cfg["color_seg_hsv_weight"]
        color_seg_dist_thresh = cfg["color_seg_dist_thresh"]
        color_seg_min_bg_dist = cfg["color_seg_min_bg_dist"]
        min_contour_area = cfg["min_contour_area"]
        contour_dist_thresh = cfg["contour_dist_thresh"]

        # foreground masking
        binary_im = color_im.foreground_mask(
            foreground_mask_tolerance, bgmodel=bgmodel
        )
        binary_im = binary_im.prune_contours(
            area_thresh=min_contour_area, dist_thresh=contour_dist_thresh
        )
        if binary_im is None:
            return None, None, None

        color_im = color_im.mask_binary(binary_im)

        # kmeans segmentation
        segment_im = color_im.segment_kmeans(
            color_seg_rgb_weight,
            color_seg_num_clusters,
            hue_weight=color_seg_hsv_weight,
        )

        # keep the segment that is farthest from the background
        bg_dists = []
        hsv_bgmodel = 255 * np.array(
            colorsys.rgb_to_hsv(
                float(bgmodel[0]) / 255,
                float(bgmodel[1]) / 255,
                float(bgmodel[2]) / 255,
            )
        )
        hsv_bgmodel = np.r_[
            color_seg_rgb_weight * np.array(bgmodel),
            color_seg_hsv_weight * hsv_bgmodel[:1],
        ]

        for k in range(segment_im.num_segments - 1):
            seg_mask = segment_im.segment_mask(k)
            color_im_segment = color_im.mask_binary(seg_mask)
            color_im_segment_data = color_im_segment.nonzero_data()
            color_im_segment_data = np.c_[
                color_seg_rgb_weight * color_im_segment_data,
                color_seg_hsv_weight
                * color_im_segment.nonzero_hsv_data()[:, :1],
            ]

            # take the median distance from the background
            bg_dist = np.median(
                np.linalg.norm(color_im_segment_data - hsv_bgmodel, axis=1)
            )
            if vis_segmentation:
                logging.info("BG Dist for segment %d: %.4f" % (k, bg_dist))
            bg_dists.append(bg_dist)

        # sort by distance
        dists_and_indices = zip(np.arange(len(bg_dists)), bg_dists)
        dists_and_indices.sort(key=lambda x: x[1], reverse=True)

        # mask out the segment in the binary image
        if (
            color_seg_num_clusters > 1
            and abs(dists_and_indices[0][1] - dists_and_indices[1][1])
            > color_seg_dist_thresh
            and dists_and_indices[1][1] < color_seg_min_bg_dist
        ):
            obj_segment = dists_and_indices[0][0]
            obj_seg_mask = segment_im.segment_mask(obj_segment)
            binary_im = binary_im.mask_binary(obj_seg_mask)
            binary_im, diff_px = binary_im.center_nonzero()
            bounding_box = Box(
                bounding_box.min_pt.astype(np.float32) - diff_px,
                bounding_box.max_pt.astype(np.float32) - diff_px,
                bounding_box.frame,
            )

        if vis_segmentation:
            plt.figure()
            plt.subplot(1, 3, 1)
            plt.imshow(color_im.data)
            plt.axis("off")
            plt.subplot(1, 3, 2)
            plt.imshow(segment_im.data)
            plt.colorbar()
            plt.axis("off")
            plt.subplot(1, 3, 3)
            plt.imshow(binary_im.data, cmap=plt.cm.gray)
            plt.axis("off")
            plt.show()

        return binary_im, segment_im, bounding_box

    def detect(
        self,
        color_im,
        depth_im,
        cfg,
        camera_intr=None,
        T_camera_world=None,
        vis_foreground=False,
        vis_segmentation=False,
        segmask=None,
    ):
        """
        Detects all relevant objects in an rgbd image pair using
        foreground masking.

        Parameters
        ----------
        color_im : :obj:`ColorImage`
            color image for detection
        depth_im : :obj:`DepthImage`
            depth image for detection (corresponds to color image)
        cfg : :obj:`YamlConfig`
            parameters of detection function
        camera_intr : :obj:`CameraIntrinsics`
            intrinsics of the camera
        T_camera_world : :obj:`autolab_core.RigidTransform`
            registration of the camera to world frame
        segmask : :obj:`BinaryImage`
            optional segmask of invalid pixels

        Returns
        ------
        :obj:`list` of :obj:`RgbdDetection`
            all detections in the image
        """
        # read params
        foreground_mask_tolerance = cfg["foreground_mask_tolerance"]
        min_contour_area = cfg["min_contour_area"]
        max_contour_area = cfg["max_contour_area"]
        min_box_area = cfg["min_box_area"]
        max_box_area = cfg["max_box_area"]
        box_padding_px = cfg["box_padding_px"]
        crop_height = cfg["image_height"]
        crop_width = cfg["image_width"]
        depth_grad_thresh = cfg["depth_grad_thresh"]

        w = cfg["filter_dim"]

        half_crop_height = float(crop_height) / 2
        half_crop_width = float(crop_width) / 2
        half_crop_dims = np.array([half_crop_height, half_crop_width])

        fill_depth = np.max(depth_im.data[depth_im.data > 0])

        kinect2_denoising = False
        if "kinect2_denoising" in cfg.keys() and cfg["kinect2_denoising"]:
            kinect2_denoising = True
            depth_offset = cfg["kinect2_noise_offset"]

        # mask image using background detection
        bgmodel = color_im.background_model()
        binary_im = color_im.foreground_mask(
            foreground_mask_tolerance, bgmodel=bgmodel
        )

        # filter the image
        y, x = np.ogrid[-w / 2 + 1 : w / 2 + 1, -w / 2 + 1 : w / 2 + 1]
        mask = x * x + y * y <= w / 2 * w / 2
        filter_struct = np.zeros([w, w]).astype(np.uint8)
        filter_struct[mask] = 1
        binary_im_filtered = binary_im.apply(
            snm.grey_closing, structure=filter_struct
        )

        # find all contours
        contours = binary_im_filtered.find_contours(
            min_area=min_contour_area, max_area=max_contour_area
        )

        if vis_foreground:
            plt.figure()
            plt.subplot(1, 2, 1)
            plt.imshow(color_im.data)
            plt.axis("off")
            plt.subplot(1, 2, 2)
            plt.imshow(binary_im_filtered.data, cmap=plt.cm.gray)
            plt.axis("off")
            plt.show()

        # threshold gradients of depth
        depth_im = depth_im.threshold_gradients(depth_grad_thresh)

        # convert contours to detections
        detections = []
        for contour in contours:
            orig_box = contour.bounding_box
            if orig_box.area > min_box_area and orig_box.area < max_box_area:
                # convert orig bounding box to query bounding box
                min_pt = orig_box.center - half_crop_dims
                max_pt = orig_box.center + half_crop_dims
                query_box = Box(min_pt, max_pt, frame=orig_box.frame)

                # segment color to get refined detection
                color_thumbnail = color_im.crop(
                    query_box.height,
                    query_box.width,
                    query_box.ci,
                    query_box.cj,
                )
                (
                    binary_thumbnail,
                    segment_thumbnail,
                    query_box,
                ) = self._segment_color(
                    color_thumbnail,
                    query_box,
                    bgmodel,
                    cfg,
                    vis_segmentation=vis_segmentation,
                )
                if binary_thumbnail is None:
                    continue
            else:
                # otherwise take original bounding box
                query_box = Box(
                    contour.bounding_box.min_pt - box_padding_px,
                    contour.bounding_box.max_pt + box_padding_px,
                    frame=contour.bounding_box.frame,
                )

                binary_thumbnail = binary_im_filtered.crop(
                    query_box.height,
                    query_box.width,
                    query_box.ci,
                    query_box.cj,
                )

            # crop to get thumbnails
            color_thumbnail = color_im.crop(
                query_box.height, query_box.width, query_box.ci, query_box.cj
            )
            depth_thumbnail = depth_im.crop(
                query_box.height, query_box.width, query_box.ci, query_box.cj
            )
            thumbnail_intr = camera_intr
            if camera_intr is not None:
                thumbnail_intr = camera_intr.crop(
                    query_box.height,
                    query_box.width,
                    query_box.ci,
                    query_box.cj,
                )

            # fix depth thumbnail
            depth_thumbnail = depth_thumbnail.replace_zeros(fill_depth)
            if kinect2_denoising:
                depth_data = depth_thumbnail.data
                binary_mask_data = binary_thumbnail.data
                depth_mask_data = depth_thumbnail.mask_binary(
                    binary_thumbnail
                ).data
                depth_mask_data += depth_offset
                depth_data[binary_mask_data > 0] = depth_mask_data[
                    binary_mask_data > 0
                ]
                depth_thumbnail = DepthImage(depth_data, depth_thumbnail.frame)

            # append to detections
            detections.append(
                RgbdDetection(
                    color_thumbnail,
                    depth_thumbnail,
                    query_box,
                    binary_thumbnail=binary_thumbnail,
                    contour=contour,
                    camera_intr=thumbnail_intr,
                )
            )

        return detections


class PointCloudBoxDetector(RgbdDetector):
    """Detect by removing all points in a point cloud that are outside of
    a given 3D bounding box.
    Converts all detections within a specified area into query images
    for a cnn. Optionally resegements the images using KMeans to remove
    spurious background pixels.
    """

    def detect(
        self,
        color_im,
        depth_im,
        cfg,
        camera_intr,
        T_camera_world,
        vis_foreground=False,
        vis_segmentation=False,
        segmask=None,
    ):
        """Detects all relevant objects in an rgbd image pair using foreground masking.

        Parameters
        ----------
        color_im : :obj:`ColorImage`
            color image for detection
        depth_im : :obj:`DepthImage`
            depth image for detection (corresponds to color image)
        cfg : :obj:`YamlConfig`
            parameters of detection function
        camera_intr : :obj:`CameraIntrinsics`
            intrinsics of the camera
        T_camera_world : :obj:`autolab_core.RigidTransform`
            registration of the camera to world frame
        segmask : :obj:`BinaryImage`
            optional segmask of invalid pixels

        Returns
        -------
        :obj:`list` of :obj:`RgbdDetection`
            all detections in the image
        """
        # read params
        min_pt_box = np.array(cfg["min_pt"])
        max_pt_box = np.array(cfg["max_pt"])
        min_contour_area = cfg["min_contour_area"]
        max_contour_area = cfg["max_contour_area"]
        min_box_area = cfg["min_box_area"]
        max_box_area = cfg["max_box_area"]
        box_padding_px = cfg["box_padding_px"]
        crop_height = cfg["image_height"]
        crop_width = cfg["image_width"]
        point_cloud_mask_only = cfg["point_cloud_mask_only"]

        w = cfg["filter_dim"]

        half_crop_height = float(crop_height) / 2
        half_crop_width = float(crop_width) / 2
        half_crop_dims = np.array([half_crop_height, half_crop_width])

        fill_depth = np.max(depth_im.data[depth_im.data > 0])

        kinect2_denoising = False
        if "kinect2_denoising" in cfg.keys() and cfg["kinect2_denoising"]:
            kinect2_denoising = True
            depth_offset = cfg["kinect2_noise_offset"]

        box = Box(min_pt_box, max_pt_box, "world")

        # project into 3D
        point_cloud_cam = camera_intr.deproject(depth_im)
        point_cloud_world = T_camera_world * point_cloud_cam
        seg_point_cloud_world, _ = point_cloud_world.box_mask(box)
        seg_point_cloud_cam = T_camera_world.inverse() * seg_point_cloud_world
        depth_im_seg = camera_intr.project_to_image(seg_point_cloud_cam)

        # mask image using background detection
        binary_im = depth_im_seg.to_binary()
        if segmask is not None:
            binary_im = binary_im.mask_binary(segmask.inverse())

        # filter the image
        y, x = np.ogrid[-w / 2 + 1 : w / 2 + 1, -w / 2 + 1 : w / 2 + 1]
        mask = x * x + y * y <= w / 2 * w / 2
        filter_struct = np.zeros([w, w]).astype(np.uint8)
        filter_struct[mask] = 1
        binary_im_filtered_data = snm.binary_dilation(
            binary_im.data, structure=filter_struct
        )
        binary_im_filtered = BinaryImage(
            binary_im_filtered_data.astype(np.uint8),
            frame=binary_im.frame,
            threshold=0,
        )

        # find all contours
        contours = binary_im_filtered.find_contours(
            min_area=min_contour_area, max_area=max_contour_area
        )

        if vis_foreground:
            plt.figure()
            plt.subplot(1, 3, 1)
            plt.imshow(color_im.data)
            plt.imshow(segmask.data, cmap=plt.cm.gray)
            plt.axis("off")
            plt.subplot(1, 3, 2)
            plt.imshow(binary_im.data, cmap=plt.cm.gray)
            plt.axis("off")
            plt.subplot(1, 3, 3)
            plt.imshow(binary_im_filtered.data, cmap=plt.cm.gray)
            plt.axis("off")
            plt.show()

        # switch to just return the mean of nonzero_px
        if point_cloud_mask_only == 1:
            center_px = np.mean(binary_im_filtered.nonzero_pixels(), axis=0)
            ci = center_px[0]
            cj = center_px[1]
            binary_thumbnail = binary_im_filtered.crop(
                crop_height, crop_width, ci, cj
            )
            color_thumbnail = color_im.crop(crop_height, crop_width, ci, cj)
            depth_thumbnail = depth_im.crop(crop_height, crop_width, ci, cj)
            thumbnail_intr = camera_intr
            if camera_intr is not None:
                thumbnail_intr = camera_intr.crop(
                    crop_height, crop_width, ci, cj
                )

            query_box = Box(
                center_px - half_crop_dims, center_px + half_crop_dims
            )
            return [
                RgbdDetection(
                    color_thumbnail,
                    depth_thumbnail,
                    query_box,
                    binary_thumbnail=binary_thumbnail,
                    contour=None,
                    camera_intr=thumbnail_intr,
                )
            ]

        # convert contours to detections
        detections = []
        for i, contour in enumerate(contours):
            orig_box = contour.bounding_box
            logging.debug("Orig box %d area: %.3f" % (i, orig_box.area))
            if orig_box.area > min_box_area and orig_box.area < max_box_area:
                # convert orig bounding box to query bounding box
                min_pt = orig_box.center - half_crop_dims
                max_pt = orig_box.center + half_crop_dims
                query_box = Box(min_pt, max_pt, frame=orig_box.frame)

                # segment color to get refined detection
                contour_mask = binary_im_filtered.contour_mask(contour)
                binary_thumbnail = contour_mask.crop(
                    query_box.height,
                    query_box.width,
                    query_box.ci,
                    query_box.cj,
                )

            else:
                # otherwise take original bounding box
                query_box = Box(
                    contour.bounding_box.min_pt - box_padding_px,
                    contour.bounding_box.max_pt + box_padding_px,
                    frame=contour.bounding_box.frame,
                )

                binary_thumbnail = binary_im_filtered.crop(
                    query_box.height,
                    query_box.width,
                    query_box.ci,
                    query_box.cj,
                )

            # crop to get thumbnails
            color_thumbnail = color_im.crop(
                query_box.height, query_box.width, query_box.ci, query_box.cj
            )
            depth_thumbnail = depth_im.crop(
                query_box.height, query_box.width, query_box.ci, query_box.cj
            )
            thumbnail_intr = camera_intr
            if camera_intr is not None:
                thumbnail_intr = camera_intr.crop(
                    query_box.height,
                    query_box.width,
                    query_box.ci,
                    query_box.cj,
                )

            # fix depth thumbnail
            depth_thumbnail = depth_thumbnail.replace_zeros(fill_depth)
            if kinect2_denoising:
                depth_data = depth_thumbnail.data
                binary_mask_data = binary_thumbnail.data
                depth_mask_data = depth_thumbnail.mask_binary(
                    binary_thumbnail
                ).data
                depth_mask_data += depth_offset
                depth_data[binary_mask_data > 0] = depth_mask_data[
                    binary_mask_data > 0
                ]
                depth_thumbnail = DepthImage(depth_data, depth_thumbnail.frame)

            # append to detections
            detections.append(
                RgbdDetection(
                    color_thumbnail,
                    depth_thumbnail,
                    query_box,
                    binary_thumbnail=binary_thumbnail,
                    contour=contour,
                    camera_intr=thumbnail_intr,
                )
            )

        return detections


class RgbdDetectorFactory:
    """Factory class for detectors."""

    @staticmethod
    def detector(detector_type):
        """Returns a detector of the specified type."""
        if detector_type == "point_cloud_box":
            return PointCloudBoxDetector()
        elif detector_type == "rgbd_foreground_mask_query":
            return RgbdForegroundMaskQueryImageDetector()
        elif detector_type == "rgbd_foreground_mask":
            return RgbdForegroundMaskDetector()
        raise ValueError("Detector type %s not understood" % (detector_type))



"""
Class to handle experiment logging.
Authors: Jeff, Jacky
"""
from abc import ABCMeta, abstractmethod
import os
import shutil
import subprocess
import logging

from .csv_model import CSVModel
from .utils import gen_experiment_id


class ExperimentLogger:
    """Abstract class for experiment logging.

    Experiments are logged to CSV files, which are encapsulated with the
    :obj:`CSVModel` class.
    """

    __metaclass__ = ABCMeta

    _MASTER_RECORD_FILENAME = "experiment_record.csv"

    def __init__(
        self,
        experiment_root_path,
        experiment_tag="experiment",
        log_to_file=True,
        sub_experiment_dirs=True,
    ):
        """Initialize an ExperimentLogger.

        Parameters
        ----------
        experiment_root_path : :obj:`str`
            The root directory in which to save experiment files.
        experiment_tag : :obj:`str`
            The tag to use when prefixing new experiments
        log_to_file : bool, optional
            Default: True
            If True will log all logging statements to a log file
        sub_experiment_dirs : bool, optional
            Defautl: True
            If True will make sub directories corresponding to generated
            experiment name
        """
        self.experiment_root_path = experiment_root_path

        # open the master record
        self.master_record_filepath = os.path.join(
            self.experiment_root_path, ExperimentLogger._MASTER_RECORD_FILENAME
        )
        self.master_record = CSVModel.get_or_create(
            self.master_record_filepath, self.experiment_meta_headers
        )

        # add new experiment to the master record
        self.id = ExperimentLogger.gen_experiment_ref(experiment_tag)
        self._master_record_uid = self.master_record.insert(
            self.experiment_meta_data
        )

        # make experiment output dir
        if sub_experiment_dirs:
            self.experiment_path = os.path.join(
                self.experiment_root_path, self.id
            )
            if not os.path.exists(self.experiment_path):
                os.makedirs(self.experiment_path)
        else:
            self.experiment_path = self.experiment_root_path

        if log_to_file:
            # redirect logging statements to a file
            if not sub_experiment_dirs:
                self.log_path = os.path.join(self.experiment_root_path, "logs")
            else:
                self.log_path = self.experiment_path
            if not os.path.exists(self.log_path):
                os.makedirs(self.log_path)
            experiment_log = os.path.join(self.log_path, "%s.log" % (self.id))
            formatter = logging.Formatter(
                "%(asctime)s %(levelname)s: %(message)s"
            )
            hdlr = logging.FileHandler(experiment_log)
            hdlr.setFormatter(formatter)
            logging.getLogger().addHandler(hdlr)

        # internal dir struct
        self._dirs = {}

    @staticmethod
    def gen_experiment_ref(experiment_tag, n=10):
        """Generate a random string for naming.

        Parameters
        ----------
        experiment_tag : :obj:`str`
            tag to prefix name with
        n : int
            number of random chars to use

        Returns
        -------
        :obj:`str`
            string experiment ref
        """
        experiment_id = gen_experiment_id(n=n)
        return "{0}_{1}".format(experiment_tag, experiment_id)

    def update_master_record(self, data):
        """Update a row of the experimental master record CSV with the given data.

        Parameters
        ----------
        uid : int
            The UID of the row to update.

        data : :obj:`dict`
            A dictionary mapping keys (header strings) to values, which
            represents the new row.
        """
        self.master_record.update_by_uid(self._master_record_uid, data)

    @abstractmethod
    def experiment_meta_headers(self):
        """Returns list of two-tuples of header names and types of meta
        information for the experiments

        Returns
        -------
        :obj:`tuple`
            The metadata for this experiment.
        """
        pass

    @abstractmethod
    def experiment_meta_data(self):
        """Returns the dict of header names and value of meta information for
        the experiments

        Returns
        -------
        :obj:`dict`
            The metadata for this experiment.
        """
        pass

    @property
    def dirs(self):
        return self._dirs.copy()

    def construct_internal_dirs(self, dirs, realize=False):
        cur_dir = self._dirs
        for dir in dirs:
            if dir not in cur_dir:
                cur_dir[dir] = {}
            cur_dir = cur_dir[dir]
        if realize:
            self._realize_dirs(dirs)

    def construct_internal_dirs_group(self, group_dirs):
        for dirs in group_dirs:
            self.construct_internal_dirs(dirs)

    def has_internal_dirs(self, dirs):
        cur_dir = self.dirs
        for dir in dirs:
            if dir not in cur_dir:
                return False
            cur_dir = cur_dir[dir]
        return True

    def dirs_to_path(self, dirs):
        rel_path = "/".join(dirs)
        abs_path = os.path.join(self.experiment_path, rel_path)
        return abs_path

    def _realize_dirs(self, dirs):
        if not self.has_internal_dirs(dirs):
            raise Exception(
                "Directory has not been constructed internally! {0}".format(
                    dirs
                )
            )
        abs_path = self.dirs_to_path(dirs)
        if not os.path.exists(abs_path):
            os.makedirs(abs_path)
        return abs_path

    def remove_dirs(self, dirs):
        if not self.has_internal_dirs(dirs):
            raise Exception(
                "Directory has not been construted internally! {0}".format(
                    dirs
                )
            )

        path = self.dirs_to_path(dirs)
        if os.path.exists(path):
            subprocess.call(["trash", "-r", path])

        # remove the deepest node
        cur_dir = self.dirs
        for dir in dirs[:-1]:
            cur_dir = cur_dir[dir]
        cur_dir.pop(dirs[-1])

        for i in range(len(dirs) - 1):
            cur_dir = self._dirs
            depth = len(dirs) - i - 2
            for j in range(depth):
                cur_dir = cur_dir[dirs[j]]

            dir_to_remove = dirs[depth]
            if not cur_dir[dir_to_remove]:
                cur_dir.pop(dir_to_remove)
            else:
                break

    def copy_to_dir(self, src_file_path, target_dirs):
        abs_path = self._realize_dirs(target_dirs)
        basename = os.path.basename(src_file_path)
        target_file_path = os.path.join(abs_path, basename)

        logging.debug(
            "Copying {0} to {1}".format(src_file_path, target_file_path)
        )
        shutil.copyfile(src_file_path, target_file_path)

    def copy_dirs(self, src_dirs_path, target_dirs):
        if not self.has_internal_dirs(target_dirs):
            raise Exception(
                "Directory has not been constructed internally! {0}".format(
                    target_dirs
                )
            )

        target_dirs_path = self.dirs_to_path(target_dirs)
        if os.path.exists(target_dirs_path):
            if len(os.listdir(target_dirs_path)) > 0:
                raise Exception(
                    "Target path for copying directories is not empty! "
                    "Got: {target_dirs_path}"
                )
            else:
                os.rmdir(target_dirs_path)
        shutil.copytree(src_dirs_path, target_dirs_path)

    @staticmethod
    def pretty_str_time(dt):
        return "{0}_{1}_{2}_{3}:{4}".format(
            dt.year, dt.month, dt.day, dt.hour, dt.minute
        )



from .version import __version__
from .csv_model import CSVModel
from .dual_quaternion import DualQuaternion
from .experiment_logger import ExperimentLogger
from .json_serialization import dump, load
from .points import BagOfPoints, BagOfVectors, Point, Direction, Plane3D
from .points import (
    PointCloud,
    NormalCloud,
    ImageCoords,
    RgbCloud,
    RgbPointCloud,
    PointNormalCloud,
)
from .primitives import Box, Contour
from .rigid_transformations import RigidTransform, SimilarityTransform
from .utils import (
    gen_experiment_id,
    histogram,
    skew,
    deskew,
    pretty_str_time,
    filenames,
    sph2cart,
    cart2sph,
    is_positive_definite,
    is_positive_semi_definite,
    keyboard_input,
)
from .yaml_config import YamlConfig
from .dist_metrics import abs_angle_diff, DistMetrics
from .random_variables import (
    RandomVariable,
    BernoulliRV,
    GaussianRV,
    ArtificialRV,
    ArtificialSingleRV,
    GaussianRigidTransformRandomVariable,
    IsotropicGaussianRigidTransformRandomVariable,
)
from .completer import Completer
from .learning_analysis import (
    ConfusionMatrix,
    ClassificationResult,
    BinaryClassificationResult,
    RegressionResult,
)
from .tensor_dataset import Tensor, TensorDatapoint, TensorDataset
from .logger import Logger
from .data_stream_syncer import DataStreamSyncer
from .data_stream_recorder import DataStreamRecorder


from .features import (
    Feature,
    LocalFeature,
    GlobalFeature,
    SHOTFeature,
    MVCNNFeature,
    BagOfFeatures,
)

from .feature_matcher import (
    Correspondences,
    NormalCorrespondences,
    FeatureMatcher,
    RawDistanceFeatureMatcher,
    PointToPlaneFeatureMatcher,
)
from .image import (
    Image,
    ColorImage,
    DepthImage,
    IrImage,
    GrayscaleImage,
    RgbdImage,
    GdImage,
    SegmentationImage,
    BinaryImage,
    PointCloudImage,
    NormalCloudImage,
)
from .chessboard_registration import (
    ChessboardRegistrationResult,
    CameraChessboardRegistration,
)
from .point_registration import (
    RegistrationResult,
    IterativeRegistrationSolver,
    PointToPlaneICPSolver,
)
from .detector import (
    RgbdDetection,
    RgbdDetector,
    RgbdForegroundMaskDetector,
    RgbdForegroundMaskQueryImageDetector,
    PointCloudBoxDetector,
    RgbdDetectorFactory,
)
from .camera_intrinsics import CameraIntrinsics



"""
Setup of core python codebase
Author: Jeff Mahler
"""
import os
from setuptools import setup

requirements = [
    "numpy",
    "scipy",
    "scikit-image",
    "scikit-learn",
    "ruamel.yaml",
    "matplotlib",
    "multiprocess",
    "setproctitle",
    "opencv-python",
    "Pillow",
    "joblib",
    "colorlog",
    "pyreadline; platform_system=='Windows'",
]

# load __version__ without importing anything
version_file = os.path.join(
    os.path.dirname(__file__), "autolab_core/version.py"
)
with open(version_file, "r") as f:
    # use eval to get a clean string of version from file
    __version__ = eval(f.read().strip().split("=")[-1])

setup(
    name="autolab_core",
    version=__version__,
    description="Core utilities for the Berkeley AutoLab",
    long_description=(
        "Core utilities for the Berkeley AutoLab. "
        "Includes rigid transformations, loggers, and 3D data wrappers."
    ),
    author="Jeff Mahler",
    author_email="jmahler@berkeley.edu",
    maintainer="Mike Danielczuk",
    maintainer_email="mdanielczuk@berkeley.edu",
    license="Apache Software License",
    url="https://github.com/BerkeleyAutomation/autolab_core",
    keywords="robotics grasping transformations",
    classifiers=[
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Natural Language :: English",
        "Topic :: Scientific/Engineering",
    ],
    packages=["autolab_core"],
    install_requires=requirements,
    extras_require={
        "docs": ["sphinx", "sphinxcontrib-napoleon", "sphinx_rtd_theme"],
        "ros": ["rospkg", "catkin_pkg", "empy"],
    },
)



"""
Generic Random Variable wrapper classes
Author: Jeff Mahler
"""
from abc import ABCMeta, abstractmethod

import numpy as np
import scipy.stats

from .rigid_transformations import RigidTransform
from .utils import skew, is_positive_semi_definite


class RandomVariable(object):
    """Abstract base class for random variables."""

    __metaclass__ = ABCMeta

    def __init__(self, num_prealloc_samples=0):
        """Initialize a random variable with optional pre-sampling.

        Parameters
        ----------
        num_prealloc_samples : int
            The number of samples to pre-allocate.
        """
        self.num_prealloc_samples_ = num_prealloc_samples
        if self.num_prealloc_samples_ > 0:
            self._preallocate_samples()

    def _preallocate_samples(self):
        """Preallocate samples for faster adaptive sampling."""
        self.prealloc_samples_ = []
        for _ in range(self.num_prealloc_samples_):
            self.prealloc_samples_.append(self.sample())

    @abstractmethod
    def sample(self, size=1):
        """Generate samples of the random variable.

        Parameters
        ----------
        size : int
            The number of samples to generate.

        Returns
        -------
        :obj:`numpy.ndarray` of float or int
            The samples of the random variable. If `size == 1`, then
            the returned value will not be wrapped in an array.
        """
        pass

    def rvs(self, size=1, iteration=1):
        """Sample the random variable, using the preallocated samples if
        possible.

        Parameters
        ----------
        size : int
            The number of samples to generate.

        iteration : int
            The location in the preallocated sample array to start sampling
            from.

        Returns
        -------
        :obj:`numpy.ndarray` of float or int
            The samples of the random variable. If `size == 1`, then
            the returned value will not be wrapped in an array.
        """
        if self.num_prealloc_samples_ > 0:
            samples = []
            for i in range(size):
                samples.append(
                    self.prealloc_samples_[
                        (iteration + i) % self.num_prealloc_samples_
                    ]
                )
            if size == 1:
                return samples[0]
            return samples
        # generate a new sample
        return self.sample(size=size)


class BernoulliRV(RandomVariable):
    """A Bernoulli random variable."""

    def __init__(self, p, *args, **kwargs):
        """Initialize a Bernoulli random variable with probability p.

        Parameters
        ----------
        p : float
            The probability that the random variable takes the value 1.
        """
        self.p = p
        super(BernoulliRV, self).__init__(*args, **kwargs)

    def sample(self, size=1):
        """Generate samples of the random variable.

        Parameters
        ----------
        size : int
            The number of samples to generate.

        Returns
        -------
        :obj:`numpy.ndarray` of int or int
            The samples of the random variable. If `size == 1`, then
            the returned value will not be wrapped in an array.
        """
        samples = scipy.stats.bernoulli.rvs(self.p, size=size)
        if size == 1:
            return samples[0]
        return samples


class GaussianRV(RandomVariable):
    """A Gaussian random variable."""

    def __init__(self, mu, sigma, *args, **kwargs):
        """Initialize a Gaussian random variable.

        Parameters
        ----------
        mu : float
            The mean of the Gaussian.

        sigma : float
            The standard deviation of the Gaussian.
        """
        self.mu = mu
        self.sigma = sigma

        super(GaussianRV, self).__init__(*args, **kwargs)

    def sample(self, size=1):
        """Generate samples of the random variable.

        Parameters
        ----------
        size : int
            The number of samples to generate.

        Returns
        -------
        :obj:`numpy.ndarray` of float
            The samples of the random variable.
        """
        samples = scipy.stats.multivariate_normal.rvs(
            self.mu, self.sigma, size=size
        )
        return samples


class ArtificialRV(RandomVariable):
    """A fake RV that deterministically returns the given object."""

    def __init__(self, obj, *args, **kwargs):
        """Initialize an artifical RV.

        Parameters
        ----------
        obj : item
            The item to always return.

        num_prealloc_samples : int
            The number of samples to pre-allocate.
        """
        self.obj_ = obj
        super(ArtificialRV, self).__init__(*args, **kwargs)

    def sample(self, size=1):
        """Generate copies of the artifical RV.

        Parameters
        ----------
        size : int
            The number of samples to generate.

        Returns
        -------
        :obj:`numpy.ndarray` of item
            The copies of the fake RV.
        """
        return np.array([self.obj_] * size)


class ArtificialSingleRV(ArtificialRV):
    """A single ArtificialRV."""

    def sample(self, size=None):
        """Generate a single copy of the artificial RV.

        Returns
        -------
        item
            The copies of the fake RV.
        """
        return self.obj_


class GaussianRigidTransformRandomVariable(RandomVariable):
    """Random variable for sampling RigidTransformations with
    a Gaussian distribution over pose variables.

    We assume no correlation between translation and rotation, so
    their values are sampled independently.

    To sample rotations, we use the method described on page 7 here:
    http://ethaneade.com/lie.pdf

    Attributes
    ----------
    mu_tra : :obj:`numpy.ndarray` of float or int
        Mean translation
    mu_rot : :obj:`numpy.ndarray` of float or int
        Mean rotation
    sigma_tra : :obj:`numpy.ndarray` of float or int
        Covariance of translation.
    sigma_rot: :obj:`numpy.ndarray` of float or int
        Covariance of rotation
    from_frame : str
    to_frame : str

    Raises
    ------
    ValueError
        If mu_rot is not a valid rotation, or if either sigma_tra or sigma_rot
        is not positive semi-definite.
    """

    def __init__(
        self,
        mu_tra=np.zeros(3),
        mu_rot=np.eye(3),
        sigma_tra=np.eye(3),
        sigma_rot=np.eye(3),
        from_frame="world",
        to_frame="world",
        *args,
        **kwargs
    ):
        if np.abs(np.linalg.det(mu_rot) - 1.0) > 1e-3:
            raise ValueError("Illegal rotation. Must have determinant == 1.0")
        if not is_positive_semi_definite(sigma_tra):
            raise ValueError(
                "Translation covariance is not positive semi-definite!"
            )
        if not is_positive_semi_definite(sigma_rot):
            raise ValueError(
                "Rotation covariance is not positive semi-definite!"
            )

        # read params
        self._mu_tra = mu_tra.copy()
        self._mu_rot = mu_rot.copy()
        self._sigma_tra = sigma_tra.copy()
        self._sigma_rot = sigma_rot.copy()

        diag_idx = np.diag_indices(3)
        self._sigma_tra[diag_idx] = np.clip(
            np.diag(self._sigma_tra), 1e-10, np.inf
        )
        self._sigma_rot[diag_idx] = np.clip(
            np.diag(self._sigma_rot), 1e-10, np.inf
        )

        self._from_frame = from_frame
        self._to_frame = to_frame

        # setup random variables
        self._t_rv = scipy.stats.multivariate_normal(
            self._mu_tra, self._sigma_tra
        )
        self._r_xi_rv = scipy.stats.multivariate_normal(
            np.zeros(3), self._sigma_rot
        )
        super(GaussianRigidTransformRandomVariable, self).__init__(
            *args, **kwargs
        )

    def sample(self, size=1):
        """Sample rigid transform random variables.

        Parameters
        ----------
        size : int
            number of sample to take

        Returns
        -------
        :obj:`list` of :obj:`RigidTransform`
            sampled rigid transformations
        """
        samples = []
        for _ in range(size):
            # sample random pose
            xi = self._r_xi_rv.rvs(size=1)
            S_xi = skew(xi)
            R_sample = scipy.linalg.expm(S_xi).dot(self._mu_rot)

            t_sample = self._t_rv.rvs(size=1)

            samples.append(
                RigidTransform(
                    rotation=R_sample,
                    translation=t_sample,
                    from_frame=self._from_frame,
                    to_frame=self._to_frame,
                )
            )

        # not a list if only 1 sample
        if size == 1 and len(samples) > 0:
            return samples[0]
        return samples


class IsotropicGaussianRigidTransformRandomVariable(
    GaussianRigidTransformRandomVariable
):
    """Random variable for sampling RigidTransformations with
    a zero-mean isotropic Gaussian distribution over pose variables.

    Attributes
    ----------
    sigma_trans : float
        variance for translation
    sigma_rot : float
        variance for rotation
    from_frame : str
    to_frame : str

    """

    def __init__(
        self,
        sigma_trans,
        sigma_rot,
        from_frame="world",
        to_frame="world",
        *args,
        **kwargs
    ):
        super(IsotropicGaussianRigidTransformRandomVariable, self).__init__(
            sigma_tra=max(1e-10, sigma_trans) * np.eye(3),
            sigma_rot=max(1e-10, sigma_rot) * np.eye(3),
            from_frame=from_frame,
            to_frame=to_frame,
            *args,
            **kwargs
        )



"""
YAML Configuration Parser
Author : Jeff Mahler
"""
import os
import ruamel.yaml as yaml
import re
from collections import OrderedDict


class YamlConfig(object):
    """Class to load a configuration file and parse it into a dictionary.

    Attributes
    ----------
    config : :obj:`dictionary`
        A dictionary that contains the contents of the configuration.
    """

    def __init__(self, filename=None):
        """Initialize a YamlConfig by loading it from the given file.

        Parameters
        ----------
        filename : :obj:`str`
            The filename of the .yaml file that contains the configuration.
        """
        self.config = {}
        if filename:
            self._load_config(filename)

    def keys(self):
        """Return the keys of the config dictionary.

        Returns
        -------
        :obj:`list` of :obj:`Object`
            A list of the keys in the config dictionary.
        """
        return self.config.keys()

    def update(self, d):
        """Update the config with a dictionary of parameters.

        Parameters
        ----------
        d : :obj:`dict`
            dictionary of parameters
        """
        self.config.update(d)

    def get(self, key, default=None):
        """Allows for get method like python dict."""
        return self.config.get(key, default)

    def __contains__(self, key):
        """Overrides 'in' operator."""
        return key in self.config.keys()

    def __getitem__(self, key):
        """Overrides the key access operator []."""
        return self.config[key]

    def __setitem__(self, key, val):
        """Overrides the keyed setting operator []."""
        self.config[key] = val

    def iteritems(self):
        """Returns iterator over config dict."""
        return self.config.iteritems()

    def save(self, filename):
        """Save a YamlConfig to disk."""
        y = yaml.YAML()
        y.dump(self.config, open(filename, "w"))

    def _load_config(self, filename):
        """Loads a yaml configuration file from the given filename.

        Parameters
        ----------
        filename : :obj:`str`
            The filename of the .yaml file that contains the configuration.
        """
        # Read entire file for metadata
        fh = open(filename, "r")
        self.file_contents = fh.read()

        # Replace !include directives with content
        config_dir = os.path.split(filename)[0]
        include_re = re.compile(r"^(.*)!include\s+(.*)$", re.MULTILINE)

        def recursive_load(matchobj, path):
            first_spacing = matchobj.group(1)
            other_spacing = first_spacing.replace("-", " ")
            fname = os.path.join(path, matchobj.group(2).rstrip())
            new_path, _ = os.path.split(fname)
            new_path = os.path.realpath(new_path)
            text = ""
            with open(fname) as f:
                text = f.read()
            text = first_spacing + text
            text = text.replace(
                "\n", "\n{}".format(other_spacing), text.count("\n") - 1
            )
            return re.sub(
                include_re, lambda m: recursive_load(m, new_path), text
            )

        # def include_repl(matchobj):
        #     first_spacing = matchobj.group(1)
        #     other_spacing = first_spacing.replace("-", " ")
        #     fname = os.path.join(config_dir, matchobj.group(2))
        #     text = ""
        #     with open(fname) as f:
        #         text = f.read()
        #     text = first_spacing + text
        #     text = text.replace(
        #         "\n", "\n{}".format(other_spacing), text.count("\n") - 1
        #     )
        #     return text

        self.file_contents = re.sub(
            include_re,
            lambda m: recursive_load(m, config_dir),
            self.file_contents,
        )
        # Read in dictionary
        self.config = self.__ordered_load(self.file_contents)

        # Convert functions of other params to true expressions
        for k in self.config.keys():
            self.config[k] = YamlConfig.__convert_key(self.config[k])

        fh.close()

        # Load core configuration
        return self.config

    @staticmethod
    def __convert_key(expression):
        """Converts keys in YAML that reference other keys."""
        if (
            type(expression) is str
            and len(expression) > 2
            and expression[1] == "!"
        ):
            expression = eval(expression[2:-1])
        return expression

    def __ordered_load(
        self, stream, Loader=yaml.Loader, object_pairs_hook=OrderedDict
    ):
        """Load an ordered dictionary from a yaml file.

        Note
        ----
        Borrowed from John Schulman.
        http://stackoverflow.com/questions/5121931/in-python-how-can-you-load-yaml-mappings-as-ordereddicts/21048064#21048064"
        """

        class OrderedLoader(Loader):
            pass

        OrderedLoader.add_constructor(
            yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG,
            lambda loader, node: object_pairs_hook(
                loader.construct_pairs(node)
            ),
        )
        return yaml.load(stream, OrderedLoader)

    def __iter__(self):
        # Converting to a `list` will have a higher memory overhead,
        # but realistically there should not be *that* many keys.
        self._keys = list(self.config.keys())
        return self

    def __next__(self):
        try:
            return self._keys.pop(0)
        except IndexError:
            raise StopIteration
